// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Text;

using Smdn.Formats;
using Smdn.Net.MessageAccessProtocols;

namespace Smdn.Net.Pop3.Protocol.Client {
  public class PopResponseReceiver : PopReceiver {
    public bool HandleAsMultiline {
      get; set;
    }

    public PopResponseReceiver(LineOrientedBufferedStream stream)
      : base(stream)
    {
      HandleAsMultiline = false;
    }

    public PopResponse ReceiveResponse()
    {
      try {
        return ParseResponce(Receive());
      }
      catch (Exception ex) {
        if (ex is PopMalformedTextException)
          throw new PopMalformedResponseException(ex.Message);
        else
          throw ex;
      }
    }

    private static readonly ByteString PositiveStatusIndicator = new ByteString("+OK");
    private static readonly ByteString NegativeStatusIndicator = new ByteString("-ERR");
    private static readonly ByteString ContinueRequestMark = new ByteString((byte)'+', Octets.SP);
    private static readonly ByteString ContinueRequestLine = new ByteString((byte)'+', Octets.CR, Octets.LF);
    private static readonly ByteString TerminationMark = new ByteString((byte)'.', Octets.CR, Octets.LF);
    private const byte Period = (byte)'.';

    private PopResponse ParseResponce(ByteString line)
    {
      // There are currently two status
      // indicators: positive ("+OK") and negative ("-ERR").  Servers MUST
      // send the "+OK" and "-ERR" in upper case.
      PopStatusIndicator? status = null;

      if (!HandleAsMultiline) {
        if (line.StartsWith(PositiveStatusIndicator))
          status = PopStatusIndicator.Positive;
        else if (line.StartsWith(NegativeStatusIndicator))
          status = PopStatusIndicator.Negative;
      }

      if (status == null) {
        if (line == TerminationMark) {
          return new PopTerminationResponse(line.ToString());
        }
        else if (line.StartsWith(ContinueRequestMark)) {
          if (HandleAsMultiline)
            return new PopFollowingResponse(line.Substring(0, line.Length - 2/*CRLF*/), line.ToString());
          else
            // continue-req     = "+" SP [base64] CRLF
            return new PopContinuationRequest(line.Substring(2, line.Length - 4/*+, SP, CRLF*/), line.ToString());
        }
        else if (line == ContinueRequestLine) {
          if (HandleAsMultiline)
            return new PopFollowingResponse(line.Substring(0, line.Length - 2/*CRLF*/), line.ToString());
          else
            // XXX: dovecot specific? '+CRLF'
            return new PopContinuationRequest(ByteString.CreateEmpty(), line.ToString());
        }
        else {
          if (HandleAsMultiline && line[0] == Period)
            // byte-stutted
            return new PopFollowingResponse(line.Substring(1, line.Length - 3/*'.' + CRLF*/), line.ToString());
          else
            return new PopFollowingResponse(line.Substring(0, line.Length - 2/*CRLF*/), line.ToString());
        }
      }
      else {
        var sepPos = line.IndexOf(Octets.SP);
        var text = (0 <= sepPos)
          ? line.Substring(sepPos + 1, line.Length - (sepPos + 1 + 2/*CRLF*/))
          : ByteString.CreateEmpty();

        return new PopStatusResponse((PopStatusIndicator)status, text, line.ToString());
      }
    }
  }
}
